#include <stdio.h>
#include <math.h>

extern int linearregression(double * xa,double * dxa,double * ya,double * dya,int nd,double * results);

int main (int argc, const char * argv[]) {
	FILE * myfile;
	int nd,ret;
	double x,dx,y,dy;
	double xa[100],ya[100],dxa[100],dya[100],results[6];
	
	nd=0;
	myfile=fopen("testlr.txt", "r");
	while(fscanf(myfile,"%lg%lg%lg%lg",&x,&dx,&y,&dy)!=EOF && nd < 100){		
		printf("x = %g   dx = %g   y = %g   dy = %g\n",x,dx,y,dy);
		xa[nd]=x;
		dxa[nd]=dx;
		ya[nd]=y;
		dya[nd++]=dy;
	}

	ret=linearregression(xa,dxa,ya,dya,nd,results);
	if(!ret){
		printf("y=a x + b; a = %g   b = %g   Err(a) = %g  Err(b) = %g  cov = %g, chi2 = %g \n",results[0],
			   results[1],results[2],results[3],results[4],results[5]);
	}else printf("linear fit failed");
    return 0;
}


int linearregression(double * xa,double * dxa,double * ya,double * dya,int nd,double * results)
/*
 this routine calculates a least squares fit on a straight line, y= a x + b,  for a maximum of 100 points
 Its parameters are:
	xa:			x values
	dxa:		errors on x values
	ya:			y values
	dya:		errors on y values
	nd:			number of data points
	results:	array of results:
					results[0] = a
					results[1] = b
					results[2] = Err(a)
					results[3] = Err(b)
					results[4] = Cov(a,b)
					results[5] = Chi^2

 the routine returns 0 upon normal completion, -1 when the fit didn't converge after 200 iterations
 */

 {
	double meanX[100];  // a maximum of 100 data points
	double meanInc[100];
	double a,b,tha,thb,z11,z12,z22,w1,w2,sx,sy,x,xm,y,del,oldchi2,newchi2,yd,stot,flamb,xminc;//nwx,nwy;
		int i,it;
		
		a=(ya[nd-1]-ya[0])/(xa[nd-1]-xa[0]);  // guesses initial values for the parameters
		b=ya[0]- a * xa[0];
		for(i=0;i<nd;i++)meanX[i]=xa[i]; // initial values of x_tilde
		oldchi2=10.e30;   // a very large initial value
		flamb=0.001;      // usual initial value for L-M coefficient
		for(it=0;it<200;it++){    // 200 iteration at maximum
			z11=z12=z22=w1=w2=0.;
			for(i=0;i<nd;i++){
				sx=dxa[i];  //these statements could be made simpler
				sy=dya[i];  //they are modified from the Obj-C version
				sx=sx*sx;
				sy=sy*sy;
				y=ya[i];
				x=xa[i];
				xm=meanX[i];
				yd=y-a*x-b;
				stot=sy+a*a*sx;
				z11+=xm*xm/stot;
				z12+=xm/stot;
				z22+=1./stot;
				w1+=yd*xm/stot;
				w2+=yd/stot;
			}
			z11*=1+flamb;    // solves 2 x 2 system
			z22*=1+flamb;
			del=z11*z22-z12*z12;
			tha=(w1*z22-w2*z12)/del;
			thb=(w2*z11-w1*z12)/del;
			for(i=0;i<nd;i++){  // calculates increments ox x_tilde_i
				sx=dxa[i];  // see above
				sy=dya[i];
				sx=sx*sx;
				sy=sy*sy;
				y=ya[i];
				x=xa[i];
				xm=meanX[i];
				yd=y-(a+tha)*xm-b-thb;
				stot=sy+a*a*sx;
				xminc=(sy*(x-xm)+sx*a*yd)/(stot *(1.+flamb));
				xm+=xminc;
				meanInc[i]=xminc;	
				meanX[i]=xm;
			}
			a+=tha;
			b+=thb;
			newchi2=0.;   // caldulates new chi2
				for(i=0;i<nd;i++){
					sx=dxa[i];
					sy=dya[i];
					sx=sx*sx;
					sy=sy*sy;
					y=ya[i];
					x=xa[i];
					xm=meanX[i];
					yd=y-a*xm-b;
					newchi2+= (x-xm)*(x-xm)/sx+ yd*yd/sy;				
				}
				if(newchi2>oldchi2){  // divergent step: multipl. lambda by 10. or reinit. to 0.001
					if(flamb>0){
						flamb*=10.;
					}else flamb=0.001;			
					a-=tha;   // restores old values of parameters and x_tilde_i
					b-=thb;
					for(i=0;i<nd;i++){
						xm=meanX[i]-meanInc[i];
						meanX[i]=xm; 
					}
					continue;
				}
				if(fabs(newchi2-oldchi2)/oldchi2<1.e-4){  // convergence goal
					if(flamb>0.){ // stars a last step without L-M coeff.
						flamb=0.;
					}else break; // fit done
				}			
				oldchi2=newchi2;
				flamb/=10.;
		}	
		
		results[0]=a;
		results[1]=b;
		results[2]=sqrt(z22/del);
		results[3]=sqrt(z11/del);
		results[4]=-z12/del;
		results[5]=newchi2;
			return (it<200 && flamb== 0. ? 0 : -1);
} 